from . import Field
from ..validators import Type


class Float(Field):
    """Field used to handle float values"""

    def __init__(self, **kwargs):
        kwargs.setdefault("validators", []).append(Type(float))
        super(Float, self).__init__(**kwargs)



from lupin import Mapper, Schema, fields as f, validators as v
from lupin.errors import InvalidDocument, InvalidLength
from models import Artist

mapper = Mapper()

mapping = mapper.register(Artist, Schema({
    "name": f.String(validators=[v.Length(max=10)]),
}))

data = {
    "name": "Leonardo da Vinci"
}

try:
    mapper.load(data, mapping, allow_partial=True)
except InvalidDocument as errors:
    error = errors[0]
    assert isinstance(error, InvalidLength)
    assert error.path == ["name"]



class Field(object):
    """Generic field that does not convert the values"""

    def __init__(self, binding=None, default=None, validators=None, read_only=False,
                 optional=False):
        """
        Args:
            binding (str): attribute name to map on object
            default (object): default value if data is absent
            validators (list): list of validators
            read_only (bool): if True the field will only be used to serialize data
            optional (bool): if True it won't raise an error if no value provided for this field
        """
        self.binding = binding
        self.default = default
        self._validators = validators or []
        self.is_read_only = read_only
        self.is_optional = optional

    def load(self, value):
        """Loads python object from JSON value

        Args:
            value (object): a value

        Returns:
            object
        """
        return value

    def dump(self, value):
        """Dump value to its JSON representation

        Args:
            value (object): a value

        Returns:
            object
        """
        return value

    def extract_attr(self, obj, key=None):
        """Get value of the `key` attribute of object.
        If field has been provided a `binding` then it will
        override `key`

        Args:
            obj (object): object to get value from
            key (str): attribute name

        Returns:
            object
        """
        key = self.binding or key
        raw_value = getattr(obj, key)
        return self.dump(raw_value)

    def validate(self, value, path):
        """Validate value againt field validators

        Args:
            value (object): value to validate
            path (list): JSON path of value
        """
        for validator in self._validators:
            validator(value, path)



import pytest
from lupin.validators import Between
from lupin.errors import InvalidRange


class TestCall(object):
    def test_raise_error_if_inferior_to_min(self):
        validator = Between(min=1, include_min=False)
        with pytest.raises(InvalidRange):
            validator(0, path=[])

        with pytest.raises(InvalidRange):
            validator(1, path=[])

    def test_raise_error_if_superior_to_max(self):
        validator = Between(max=1, include_max=False)
        with pytest.raises(InvalidRange):
            validator(1, path=[])

        with pytest.raises(InvalidRange):
            validator(2, path=[])

    def test_does_not_raise_error_if_ok(self):
        validator = Between(min=1, max=2)
        validator(1.5, path=[])
        validator(2, path=[])
        validator(1, path=[])



import datetime
from lupin import fields
import pytest


@pytest.fixture
def field():
    return fields.Date()


@pytest.fixture
def date():
    return datetime.date(2017, 10, 7)


class TestDump(object):
    def test_returns_date_as_string(self, field, date):
        assert field.dump(date) == "2017-10-07"


class TestLoad(object):
    def test_returns_a_date(self, field, date):
        assert field.load("2017-10-07") == date



from . import Field
from ..validators import Type
from ..errors import MissingPolymorphicKey, InvalidPolymorphicType
from ..utils import get_mapping


class PolymorphicList(Field):
    """Field used to handle a list containing objects
    of different types, that need to be loaded, serialized
    in different ways.

    Example:
        PolymorphicList(on="type",
                        mappings={
                            "diamond": diamond_mapping,
                            "painting": painting_mapping
                        })
    """
    def __init__(self, on, mappings, **kwargs):
        """
        Args:
            on (str): JSON key used to get the object type
            mappings (dict): mapping used for each values used for the `on` key
        """
        kwargs.setdefault("validators", []).append(Type(list))
        super(PolymorphicList, self).__init__(**kwargs)
        self._on = on
        self._mappings_by_json_value = mappings
        self._mappings_by_type = {mapping.cls: mapping for mapping in mappings.values()}

    def load(self, value):
        """Loads list of python objects from JSON list

        Args:
            value (list): JSON list

        Returns:
            list
        """
        return [self._mappings_by_json_value[item[self._on]].load(item)
                for item
                in value]

    def dump(self, value):
        """Dump list of object

        Args:
            value (list): objects to dump

        Returns:
            list
        """
        return [get_mapping(self._mappings_by_type, item).dump(item)
                for item
                in value]

    def validate(self, value, path):
        """Validate each items of list against nested mapping.

        Args:
            value (list): value to validate
            path (list): JSON path of value
        """
        super(PolymorphicList, self).validate(value, path)
        for item_index, item in enumerate(value):
            item_path = path + [str(item_index)]
            if self._on not in item:
                raise MissingPolymorphicKey(self._on, path)

            obj_type = item[self._on]
            if obj_type not in self._mappings_by_json_value:
                raise InvalidPolymorphicType(invalid_type=obj_type,
                                             supported_types=list(self._mappings_by_json_value.keys()),
                                             path=path+[self._on])

            mapping = self._mappings_by_json_value[obj_type]
            mapping.validate(item, path=item_path)



class Thief(object):
    """Thief fake class"""

    def __init__(self, first_name, last_name):
        """
        Args:
            first_name (str): thief first name
            last_name (str): thief last name
        """
        self.first_name = first_name
        self.last_name = last_name



from . import bind


class Mapping(object):
    """Association of a serializable class and a schema"""

    def __init__(self, cls, schema, factory=bind):
        """
        Args:
            cls (class): class to load and dump
            schema (Schema): a schema
            factory (callable): a factory method for instantiating objects
        """
        self.cls = cls
        self._schema = schema
        self._factory = factory

    def load(self, data, allow_partial=False):
        """Load an instance of self._cls with values contained in data

        Args:
            data (dict): JSON data
            allow_partial (bool): allow partial schema, won't raise error if missing keys

        Returns:
            object
        """
        return self._schema.load(self.cls, data, allow_partial, self._factory)

    def load_attrs(self, data, allow_partial=False):
        """Loads attributes dictionary from `data`

        Args:
            data (dict): dictionary of data
            allow_partial (bool): allow partial schema, won't raise error if missing keys

        Returns:
            dict
        """
        return self._schema.load_attrs(data, allow_partial)

    def dump(self, obj):
        """Dump object into its JSON representation

        Args:
            obj (object)

        Returns:
            dict
        """
        return self._schema.dump(obj)

    def validate(self, data, allow_partial=False, path=None):
        """Validate data with the schema.
        If path is provided it will be used as the base path for errors.

        Args:
            data (dict): data to validate
            allow_partial (bool): allow partial schema, won't raise error if missing keys
            path (list): base path for errors
        """
        self._schema.validate(data, allow_partial, path)



from ..errors import InvalidRange


class Between(object):
    """Validates that a value is contained within bounds"""
    def __init__(self, min=None, max=None, include_min=True, include_max=True):
        """
        Args:
            min (int): minimum value
            max (int): maximum value
            include_min (bool): include min value
            include_max (bool): include max value
        """
        self._min = min
        self._max = max
        self._include_min = include_min
        self._include_max = include_max

    def __call__(self, value, path):
        """Validate that value is contained is the range

        Args:
            value (object): value to validate
            path (list): error path
        """
        if self._min is not None and\
                (self._include_min and value < self._min or\
                 not self._include_min and value <= self._min):
            raise InvalidRange(value, self._min, self._max, path=path)

        if self._max is not None and\
                (self._include_max and value > self._max or\
                 not self._include_max and value >= self._max):
            raise InvalidRange(value, self._min, self._max, path=path)



from . import Field
from ..validators import Type


class Dict(Field):
    """Field used to handle dictionnaries"""

    def __init__(self, **kwargs):
        kwargs.setdefault("validators", []).append(Type(dict))
        super(Dict, self).__init__(**kwargs)



from . import Field
from ..validators import Equal


class Constant(Field):
    """Field used to display a fixed value"""

    def __init__(self, value, **kwargs):
        """
        Args:
            value (object): fixed value
        """
        kwargs.setdefault("validators", []).append(Equal(value))
        super(Constant, self).__init__(**kwargs)
        self._value = value

    def load(self, value):
        """Returns fixed value

        Args:
            value (object): ignored

        Returns:
            object
        """
        return self._value

    def dump(self, value):
        """Returns fixed value

        Args:
            value (object): ignored

        Returns:
            object
        """
        return self._value

    def extract_attr(self, obj, key=None):
        """Returns fixed value

        Args:
            obj (object): object to get value from
            key (str): attribute name

        Returns:
            object
        """
        return self._value



from . import Field
from ..validators import Type


class Object(Field):
    """Field used for nested objects.
    It uses a mapping in order to know what class to load and the schema
    used for the JSON representation.
    """

    def __init__(self, mapping, *args, **kwargs):
        """
        Args:
            mapping (Mapping): mapping of nested object
        """
        kwargs.setdefault("validators", []).append(Type(dict))
        super(Object, self).__init__(*args, **kwargs)
        self._mapping = mapping

    def load(self, value):
        """Loads python object from JSON object

        Args:
            value (dict): nested JSON object

        Returns:
            object
        """
        return self._mapping.load(value)

    def dump(self, value):
        """Dump object into its JSON representation

        Args:
            value (object): an object

        Returns:
            dict
        """
        return self._mapping.dump(value)

    def validate(self, value, path):
        """Validate value against mapping validators.

        Args:
            value (list): value to validate
            path (list): JSON path of value
        """
        super(Object, self).validate(value, path)
        self._mapping.validate(value, path)



import pytest

from lupin import Mapper, Object
from lupin.errors import InvalidDocument
from tests.fixtures import Thief


@pytest.fixture
def mapping(thief_schema):
    mapper = Mapper()
    return mapper.register(Thief, thief_schema)


@pytest.fixture
def field(mapping):
    return Object(mapping)


class TestLoad(object):
    def test_returns_a_thief(self, field, thief_data):
        thief = field.load(thief_data)
        assert isinstance(thief, Thief)


class TestDump(object):
    def test_returns_thief_data(self, field, thief, thief_data):
        data = field.dump(thief)
        assert data == thief_data


class TestValidate(object):
    def test_does_nothing_if_valid(self, field, thief_data):
        field.validate(thief_data, [])

    def test_raises_error_if_invalid_data(self, field, thief_data):
        thief_data["firstName"] = 46
        with pytest.raises(InvalidDocument):
            field.validate(thief_data, [])



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
    Setup file for lupin.

    This file was generated with PyScaffold 2.5.8, a tool that easily
    puts up a scaffold for your new Python project. Learn more under:
    http://pyscaffold.readthedocs.org/
"""

import sys
from setuptools import setup


try:
    import pypandoc
    long_description = pypandoc.convert("README.md", "rst")
    with open("README.rst", "w") as f:
        f.write(long_description)
except (ImportError, OSError):
    with open("README.md") as f:
        long_description = f.read()


def setup_package():
    needs_sphinx = {"build_sphinx", "upload_docs"}.intersection(sys.argv)
    sphinx = ["sphinx"] if needs_sphinx else []
    setup(setup_requires=["six", "pyscaffold>=2.5a0,<2.6a0"] + sphinx,
          use_pyscaffold=True,
          long_description=long_description)


if __name__ == "__main__":
    setup_package()



from . import Field
from ..validators import Type


class String(Field):
    """Field used to handle string values"""

    def __init__(self, **kwargs):
        kwargs.setdefault("validators", []).append(Type(str))
        super(String, self).__init__(**kwargs)
